//==============================================================================
//   FILE : PCNDRV.C
//
//   DESC : Contains PCnet-ISA driver
//
// TARGET : 16bit Real Mode x86
//
//  TOOLS : Microsoft (R) C/C++ Optimizing Compiler Version 8.00c
//
//                $Revision: 1.0 $
//              $Name: dougp$
//              $Date: 1998/24/12 12:38:33 $
//            $Author: Chipf $
//     
//------------------------------------------------------------------------------
//  Copyright 1997 Advanced Micro Devices, Inc.                                
//                                                                             
//  This software is the property of Advanced Micro Devices, Inc  (AMD)  which 
//  specifically  grants the user the right to modify, use and distribute this 
//  software provided this notice is not removed or altered.  All other rights 
//  are reserved by AMD.                                                       
//                                                                             
//  AMD MAKES NO WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, WITH REGARD TO THIS 
//  SOFTWARE.  IN NO EVENT SHALL AMD BE LIABLE FOR INCIDENTAL OR CONSEQUENTIAL 
//  DAMAGES IN CONNECTION WITH OR ARISING FROM THE FURNISHING, PERFORMANCE, OR 
//  USE OF THIS SOFTWARE.                                                      
//                                                                             
//  So that all may benefit from your experience, please report  any  problems 
//  or suggestions about this software back to AMD.  Please include your name, 
//  company,  telephone number,  AMD product requiring support and question or 
//  problem encountered.                                                       
//                                                                             
//  Advanced Micro Devices, Inc.       Worldwide support and contact           
//  Logic Products Division            information available at:               
//  Systems Engineering                                                        
//  5204 E. Ben White Blvd.     http://www.amd.com/html/support/techsup.html   
//  Austin, TX 78741                                                           
// =============================================================================
/*                         
  pcndrive.c - 32 bit network interface driver for use with the PCI/PCNet chip.

	This driver requires the following macro definitions to be
	included in a file called "pcnlocal.h"

	These are constant values:

	PCN_MAX		Maximum number of PCI/PCNet drivers to be active.

	PCN_MAX_TX	Maximum number of transmit descriptors.  This must
			be 2^n where n = 0 to 10

	PCN_MAX_RX	Maximum number of receive descriptors.  This must
			be 2^n where n = 0 to 10

	PCN_MAX_TX_SIZE	Maximum size of transmit buffer.

	PCN_MAX_RX_SIZE	Maximum size of receive buffer.

	DEBUGx((printf spec))
		Used for debugging & diagnostic purposes.  x is 1-5 and
		is used in the following way:

		DEBUG1 - Unrecoverable system error.  Usually results in
			 a "hang"

		DEBUG2 - System failure.  The system will attempt to
			 recover.

		DEBUG3 - Typically an "ASSERT" error.  This is something      
			 that may indicate a failure or potential future
			 failure.

		DEBUG4 - Informative.

		DEBUG5 - Dump.

		The usage is the same as for printf except that an
		additional set of () is required:
			Nprintf("Number: %d\n", number);
			DEBUG5(("Number: %d\n", number));

	DEBUGIx((printf spec))
		Same as DEBUGx except for use inside an interrupt routine.
		
		
	Note that this driver does not use interrupts.  It operates by
	examining the own bits in the rx and tx descriptors.  As a result,
	it relies on the Process routing being called on a regular basis.
	It would also be possible to call process from an interrupt.
*/


// include files 

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <memory.h>

#include "amdrive.h"
#include "pcn.h"
#include "pcndrv.h"
#include "am186cc.h"
#include "stdincs.h"
#include "led.h"

// Some local validation

#define Nprintf printf

#if PCN_MAX_TX_SIZE > 4095
#error PCN_MAX_TX_SIZE is too large
#endif

#if PCN_MAX_RX_SIZE > 4095
#error PCN_MAX_RX_SIZE is too large
#endif

// we allocate an oversized block and then align it in Init.
// char mem_block[(sizeof(PcnInfo_s)*PCN_MAX)+PCN_MAX_RX*8];
// The initializtion routine uses these constants for PCnet memory allocation.

#ifdef DRAM                     
#ifdef MCS3
#define SRAM_BASE 	0xb0000     // these are used for MSC3
#define SRAM_BASEL 	0xb0000000
#define SRAM_SIZE 	0x10000
#else
#define SRAM_BASE 	0x80000		// these are used for MSC0
#define SRAM_BASEL 	0x80000000
#define SRAM_SIZE 	0x10000   
#endif
#else
#ifdef SMALLSRAM                // use these if SRAM is 24k
#define SRAM_BASE	0x8000
#define SRAM_BASEL 	0x08000000
#define SRAM_SIZE 	0x6000
#else
#define SRAM_BASE 	0x20000		// for 64k SRAM
#define SRAM_BASEL 	0x20000000
#define SRAM_SIZE 	0x10000
#endif
#endif

struct PcnInfo_s *pcnInfo;
//PcnInitBlock_s	initBlk;
PcnInitBlock_s	 *initBlkp;
//#pragma BSS ;

static char dbuff[128];	// Debug buffer 

//#define MAX_SPEED 1
//#define WARP_SPEED 1

int tx_loops;			// how many times we've transmitted
int tx_done;			// flag saying TX interrupt is done


// Local prototypes 
AMERROR pcnInitializeDescriptors(int unitNumber);

/*********************************************************************

 IntHandler - Ethernet interrupt handler

 This interrupt handler deals with packet receive and transmit 
 interrupts and checks for error flags in the CSR4 & CSR0 registers

*********************************************************************/

void interrupt IntHandler(void)
{
	int		ioport;
	PcnInfo_s *pinfo;
	int unitNumber;
	int	temp_rap;
	VU16 csr_val,temp_csr4;
	
	unitNumber = 0;	// safe assumption for this test code

	pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);

// this routine assumes that the RAP has been left pointing at CSR0
    ioport = pinfo->ioport;
    
	temp_rap = _inpw(ioport+RAP);		// save the current rap 
	
	_outpw( ioport+RAP,0);				// point to CSR0 
	csr_val = _inpw(ioport+RDP);


	if (csr_val & (CSR0_TINT|CSR0_RINT|CSR0_ERR)==0) // is this a CSR4 error
	{
		_outpw( ioport+RAP,4);			// point to CSR4
		temp_csr4 = _inpw(ioport+RDP);
		DEBUG4(("Inthandler error, csr0 = %x, csr4 = %x\n",csr_val,temp_csr4));
		
		if(temp_csr4 & CSR4_MFCO)
			DEBUG4(("\nMissed Frame Counter Overflow"));
		if(temp_csr4 & CSR4_JAB)
			DEBUG4(("\nJabber Error."));
		if(temp_csr4 & CSR4_RVCO)
			DEBUG4(("\nReceive Collision Counter Overflow."));
		
		_outpw( ioport+RAP,0);			// point to CSR0 
		
	}	
	if (csr_val & CSR0_ERR)				// check for errors 
	{
		PostLEDs(0xaa);
		_outpw(ioport+RDP,(csr_val&7800)|CSR0_INEA); // reset the error interrupt
		
	}
	
	if (csr_val & CSR0_TINT)			// check for transmit interrupts 
	{
	
		
	DEBUG4(("Inthandler Transmit, CSR4 = %x\n",_inpw(ioport+RDP)));

		_outpw(ioport+RDP,CSR0_TINT | CSR0_INEA); // reset the transmit interrupt
		tx_done = 1;
	}

	if (csr_val & CSR0_RINT)	// check for receive interrupts 
	{
		DEBUG4(("Inthandler Receive\n",csr_val));
		_outpw(ioport+RDP,CSR0_RINT|CSR0_INEA); // reset the receive interrupt
	}
	
	DEBUG5(("TMD1 = %x\n",pinfo->txDesc[pinfo->nextTx].tmd1));
	DEBUG5(("TMD3 = %x\n",pinfo->txDesc[pinfo->nextTx].tmd3));

	_outpw(ioport+RAP,temp_rap);	// restore the current rap 
	_outpw(EOI,EOITYPE_INT3);		// clear the interrupt controller
}

/*********************************************************************

 pcnInit - network interface initialization routine

 This routine resets, configures, and initializes the chip.

RETURNS:
	AMERROR_OK		Successful
	AMERROR_INVALID_UNIT	The unitNumber is invalid.
	AMERROR_NO_DEVICE	The ID for the PCI/PCNet card
				was not available in the configuration
				space.
	AMERROR_NO_IO		The card's IO space could not be
				programmed.
	AMERROR_CONST		One of the constant values
				(PCN_MAX_TX, or PCN_MAX_RX) was invalid.
*********************************************************************/

AMERROR
pcnInit(int unitNumber,PcnInit_s *ppcnInit)
{
	PcnInfo_s	*pinfo;
	int		ioport;
	int		tlen,rlen;
	int		ii,iseg; 
	unsigned long	faraddr;
	U16		iaddr;

// Initialize the appropriate pcnInfo structure and
// try and talk to the chip.

	if (unitNumber >= PCN_MAX) {
		DEBUG2(("Invalid unit number: %d\n", unitNumber));
		return(AMERROR_INVALID_UNIT);
	}

// install the interrupt vector for INT3, and enable the interrupt.
    *(LPVOID far *)(ITYPE_INT3*4) = &IntHandler;
    
    _outpw(EOI,EOITYPE_INT3);		// clear the interrupt controller
	_outpw(CH8CON,CHCON_PR1|CHCON_LTM);
    _outpw(IMASK,_inpw(IMASK)&(~IMASK_CH8));

// set the pcnInfo pointer to a suitably aligned address
	faraddr = (unsigned long)(SRAM_BASE);
	iseg = (int)((faraddr&0xffff0000)>>16);		// get the segment
	
	pcnInfo = (PcnInfo_s *) (SRAM_BASEL);

//	DEBUG5(("pcnInfo = %04x\n",((U32)(pcnInfo)&0xffff0000)>>16));
//	DEBUG5(("          %04x\n",pcnInfo));

	pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);

	ioport = ppcnInit->ioport;
	pinfo->ioport = ioport;

	//Am186 ES initialize the PCS Chip Select and enable Interrupt 0

//The old driver set up chip selects and PIOs here.  I have moved them to the
// main application.  They are now in a routine so they can be called.

	// Let's see if there's really a chip out there...
	
// Make sure chip is in STOP mode 
	_outpw(ioport+RAP, 0);
	_outpw(ioport+RDP, CSR0_STOP);

// See if it's really there 
	_outpw(ioport+RAP, 88);
	ii = (_inpw(ioport+RDP)&0xf000)>>12; // get the 4 low bits of the id
	_outpw(ioport+RAP, 89);
	ii += (_inpw(ioport+RDP)&0xfff)<<4;  // get the 12 high bits of the id
	
	if (ii != 0x2261)	// unique for the PCnetISA II
	{
		DEBUG2(("Bad Chip ID, id = %x\n", ii));
		return(AMERROR_NO_IO);
	}

// Automatically pad to 64 bytes
	
	_outpw(ioport+RAP, 4);
	_outpw(ioport+RDP, 1 <<10);

// Set up the descriptors
			
	pcnInitializeDescriptors(unitNumber);
	tx_loops = 0;
        
// Set up the initialization block
	
	initBlkp = (PcnInitBlock_s *) (SRAM_BASEL + SRAM_SIZE - 0x100);
	memset(initBlkp, 0, sizeof(struct PcnInitBlock_s));
    
    // See if we should use burned-in or passed in address.
	
	if (ppcnInit->useBurned)
	{

DEBUG5(("Using burned address\n"));
		ppcnInit->eAddr[0] = initBlkp->padr[0] = _inpw(ioport)&0xff;
		ppcnInit->eAddr[1] = initBlkp->padr[1] = _inpw(ioport)>>8;
		ppcnInit->eAddr[2] = initBlkp->padr[2] = _inpw(ioport+2)&0xff;
		ppcnInit->eAddr[3] = initBlkp->padr[3] = _inpw(ioport+2)>>8;
		ppcnInit->eAddr[4] = initBlkp->padr[4] = _inpw(ioport+4)&0xff;
		ppcnInit->eAddr[5] = initBlkp->padr[5] = _inpw(ioport+4)>>8;
	}
	else
		for(ii=0;ii<6;ii++) initBlkp->padr[ii] = ppcnInit->eAddr[ii];

	for (ii=0;ii<6;ii++) pinfo->address[ii] = initBlkp->padr[ii];

	DEBUG4(("Ethernet Address %02x %02x %02x %02x %02x %02x\n",
				pinfo->address[0],
				pinfo->address[1],
				pinfo->address[2],
				pinfo->address[3],
				pinfo->address[4],
				pinfo->address[5]));
        
	
	
// Set up addresses for RX and TX descriptors 
	FTP( &pinfo->rxDesc[0],faraddr);
	initBlkp->rdral = (VU16)(0xffff & faraddr);
	initBlkp->rdrah = (VU8)((0xff0000 & faraddr)>>16);
	FTP( &pinfo->txDesc[0],faraddr);
	initBlkp->tdral = (VU16)(0xffff & faraddr);
	initBlkp->tdrah = (VU8)((0xff0000 & faraddr)>>16);

	DEBUG5(("Rdra: %02x%04x, Trda: %02x%04x\n",
		initBlkp->rdrah,initBlkp->rdral, initBlkp->tdrah,initBlkp->tdral));

// Convert PCN_MAX_nX to tlen and rlen for tmode register.

	if (xLenConv(PCN_MAX_TX, &tlen) != AMERROR_OK) {
		DEBUG2(("Invalid PCN_MAX_TX\n"));
		return((AMERROR_CONST));
	}

	if (xLenConv(PCN_MAX_RX, &rlen) != AMERROR_OK) {
		DEBUG2(("Invalid PCN_MAX_RX\n"));
		return((AMERROR_CONST));
	}

	initBlkp->rdrlen = rlen << 5;
	initBlkp->tdrlen = tlen << 5;
	
// THIS IS A VERY IMPORTANT LINE OF CODE FOR CONFIGURING THE CONTROLLER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// 0x0080 - assigned to tmode below will result in normal ethernet operation of the controller
// 			where packet address information is checked upon reception.  The controller may
//			be attached to a live network in this mode.
// 0xb084 - turn on promiscuous mode and loopback.  A loopback plug may be used to
//			return sent packets to the controller.  The controller may also be connected
//			to a live network as long as there is someone to respond to its packets in this
//			mode, but that would be an overkill since it is in loopback mode.
	
	initBlkp->tmode = 0xb084; 

// Set up pointers to the init blk and start it rolling.
	                         

	FTP( initBlkp,faraddr);
	_outpw(ioport+RAP, 1);	// CSR1 = Init block low address 
	_outpw(ioport+RDP, (VU16)(0xffff & faraddr));
	_outpw(ioport+RAP, 2);	// CSR2 - Init block upper address 
	_outpw(ioport+RDP, (VU16)((0xff0000 & faraddr)>>16));
		
	_outpw(ioport+RAP, 1);
	iaddr = _inpw(ioport+RDP);
	_outpw(ioport+RAP, 2);
	
	DEBUG5(("Init block from CSRs: %02x%04x\n",
	_inpw(ioport+RDP) & 0xff, iaddr));

	_outpw(ioport+RAP, 3);   	// set to CSR3, the interrupt masks
	_outpw(ioport+RDP, 0x5d00);	// mask off RINT

	_outpw(ioport+RAP, 4);   	// set to CSR4, other interrupt masks
	_outpw(ioport+RDP, 0x4);	// mask off txstart interrupt
#ifdef WARP_SPEED
	_outpw(ioport+RAP, 80);   	// set to CSR80
	_outpw(ioport+RDP, 0x2010);	// set for min txstart latency
								// XMSTP = 4 
#endif
	
	// Start the PCnet ISA.  It will read the init block start itself up. 

	_outpw(ioport+RAP, 80);   	// set to CSR80    
	DEBUG5(("CSR15 = %x", _inpw(ioport+RDP)));

	_outpw(ioport+RAP, 0);
	_outpw(ioport+RDP, CSR0_INIT | CSR0_STRT);

		
// Wait for it to complete 
	while ((_inpw(ioport+RDP)&0x8100)!=0x100);	// If error or finish 

	_inpw(ioport+RDP); 
	
	DEBUG4(("Init block complete: %04x\n", _inpw(ioport+RDP)));      
		
	_outpw(ioport+RAP, 0);
	_outpw(ioport+RDP, CSR0_STOP);
	_outpw(ioport+RAP, 15);
	DEBUG4(("CSR15 = %02x\n", _inpw(ioport+RDP)));
	_outpw(ioport+RAP, 2);
	DEBUG4(("CSR2 = %02x\n", _inpw(ioport+RDP)));
	_outpw(ioport+RAP, 0);
#ifdef MAX_SPEED
	_outpw(ioport+RDP, CSR0_STRT);
#else	
	_outpw(ioport+RDP, CSR0_INEA|CSR0_STRT);
#endif	
   	DEBUG5(("Init Done: CSR0 = %x\n", _inpw(ioport+RDP)));

// Wait for it to start 
	
	while ((_inpw(ioport+RDP)&0x002)!=0x2);
	ii = _inpw(ioport+RDP);
	if(ii&0x8000 == 0x8000)  // do we have an error condition?
	{
		_outpw(ioport+RDP, ii&0x7800);
	}
   	DEBUG5(("Init Done: CSR0 = %x\n", _inpw(ioport+RDP)));
   	
	return (AMERROR_OK);
}
/*********************************************************************

pcnShutdown - Stop the PCnet driver.

RETURNS:
	AMERROR_OK		Packet sent
	AMERROR_INVALID_UNIT	Invalid unit number.
	AMERROR_FULL		Can't send - output full
	AMERROR_TOO_BIG		Length to big for transmit buffer.

*********************************************************************/
	
AMERROR
pcnShutdown(int unitNumber)
{
	PcnInfo_s *pinfo;
	int  ioport;


	if (unitNumber >= PCN_MAX) {
		DEBUG2(("Invalid unit number: %d\n", unitNumber));
		return(AMERROR_INVALID_UNIT);
	}

	pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);

	_outpw(ioport+RAP, 0);
	_outpw(ioport+RDP, CSR0_STOP);
	return (AMERROR_OK);
}
/*********************************************************************

pcnOutput - Send packet out the Ethernet.

RETURNS:
	AMERROR_OK				Packet sent
	AMERROR_INVALID_UNIT	Invalid unit number.
	AMERROR_FULL			Can't send - output full
	AMERROR_TOO_BIG			Length to big for transmit buffer.

*********************************************************************/
	
AMERROR
pcnOutput(int unitNumber, U8 *txBuff, int length)
{
	int	nextTx;
	PcnInfo_s *pinfo;
	int  txLen;		// Calculate packet length 
	int  ioport;
	int  tmd1=0;
	int  i2;


#if DEBUG_LEVEL > 5
	U8 *txp;
	int  ii,jj;
	char *dptr;
#endif

	

	if (unitNumber >= PCN_MAX) {
		DEBUG2(("Invalid unit number: %d\n", unitNumber));
		return(AMERROR_INVALID_UNIT);
	}
    
    pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);

	nextTx = pinfo->nextTx;

	DEBUG5(("Transmit: %d bytes\n", length));
#ifdef MAX_SPEED
    // Poll the own bit of the next descriptor until the PCnet is done with it.
    // This will allow us to fill all of the buffers except the on the PCnet currently
    // has.  We will keep the descriptor ring full, so the PCnet is never waiting on
    // us for the next packet.  This should allow the Net186 to run at it's max
    // possible rate. 
    // 500 pps with descriptor poll interval on.
    // 5600 with MAX_SPEED off and Transmit Demand and always filling buffers
    // 9600 pps with no waiting on interrupt and static buffers
    // Got to 10600 pps with XMSTP = 16 bytes (or 4, made no difference)
    // Upping number of descriptors made no difference.
    // removed interrupt processing and got to 13000 pps!
	while(	pinfo->txDesc[nextTx].tmd1 & TMD1_OWN);

#else
	
// If owned by the adapter, then tx buffer must be full.
	
	if (pinfo->txDesc[nextTx].tmd1 & TMD1_OWN) {
		DEBUG3(("Transmit buffer full\n"));
		return(AMERROR_FULL);
	}
#endif

	if (length >= PCN_MAX_TX_SIZE) {
		DEBUG3(("TX Buffer too big: %d\n", length));
		return(AMERROR_TOO_BIG);
	}

	
// Copy the data into the correct buffer.
	
#ifdef MAX_SPEED
	if (tx_loops++ <= PCN_MAX_TX) memcpy(pinfo->txBuff[nextTx], txBuff, length);
#else
	memcpy(pinfo->txBuff[nextTx], txBuff, length);
#endif

// Dump the transmit buffer
	
#if DEBUG_LEVEL > 5
	txp = (U8 *)pinfo->txBuff[nextTx];
	for (ii = 0; ii < length; ii += 20) {
		dptr = dbuff;
		for (jj = ii; (jj < ii + 20) && (jj < length); jj++) {
			sprintf(dptr, "%02x ", txp[jj]);
			dptr += 3;
		}
		DEBUG5(("Tx: %s\n", dbuff));
	}

#endif
	
// Set descriptor ready to transmit
	
	tx_done = 0;
	_outpw(ioport+RAP, 0);
	DEBUG4(("Transmit start: CSR0 = %x\n", _inpw(ioport+RDP)));  // was debug5

	pinfo->txDesc[nextTx].tmd3 = 0;
//	if (length<64) length = 64;
	txLen = (-length) & TMD2_BCNT;
	pinfo->txDesc[nextTx].tmd2 = TMD2_ONES|txLen;
	tmd1 = pinfo->txDesc[nextTx].tmd1&0xff;
	pinfo->txDesc[nextTx].tmd1 = (tmd1|TMD1_OWN|TMD1_STP|TMD1_ENP)&(~TMD1_ADDFCS); 
        
    ioport = pinfo->ioport;     // assign chip select address to ioport

	DEBUG5(("Transmit Start: CSR0 = %x\n", _inpw(ioport+RDP)));
	DEBUG5(("TMD1 = %x\n",pinfo->txDesc[nextTx].tmd1));
	DEBUG5(("TMD3 = %x\n",pinfo->txDesc[nextTx].tmd3));
	//_outpw(ioport+RDP,inpw(ioport + RDP) | 0x00  );
	_outpw(ioport+RDP,  _inpw(ioport+RDP)| CSR0_TDMD );	// Transmit poll immediately 
	
	// printf("\nCSR3= %x\n",ip_ecr_r(3));
	//	printf("\nCSR4= %x\n",ip_ecr_r(4));
    
#ifndef MAX_SPEED  // turn on for "wait for interrupt" mode

// WAIT FOR COMPLETE 

//	while (pinfo->txDesc[nextTx].tmd1 & TMD1_OWN) ; // polled mode
	i2 = 0;
	while (!tx_done)
	{
      	i2++;
   	
	DEBUG5(("CSR4  = %x\n", _inpw(ioport+RDP)));
	DEBUG5(("CSR0  = %x\n", _inpw(ioport+RDP))); // interrupt mode
   	
	}
	DEBUG5(("Transmit done: CSR0 = %x\n", _inpw(ioport+RDP)));
	
#endif

	nextTx = (nextTx + 1) % PCN_MAX_TX;
	pinfo->nextTx = nextTx;
	return(AMERROR_OK);
} 


/*********************************************************************
pcnRProcess()

buffLength	Maximum length of received data.
rxBuff		Where to put the received data
*rxLength	Where to put the length of received data.

Check the receiver and copy a RX buffer if any available.

RETURNS:
	AMERROR_OK		No frame, no problem
	AMERROR_FRAME		Frame is received.
	AMERROR_INVALID_UNIT	Invalid unit number.
	AMERROR_TOO_BIG		Received frame too big for buffer.

*********************************************************************/

AMERROR
pcnRProcess(int unitNumber, int buffLength, U8 *rxBuff, int *prxLength)
{
	int	nextRx;
	int	ret = AMERROR_FRAME;	// Return code for CLEANUP 
	U16	rmd1;
	PcnInfo_s *pinfo;
	int	rxLength;
        int	ioport;
#if DEBUG_LEVEL > 5
	int	readTx;
	int	ii,jj;
	char	*dptr;
#endif
	if (unitNumber >= PCN_MAX) {
		DEBUG2(("Invalid unit number: %d\n", unitNumber));
		return(AMERROR_INVALID_UNIT);
	}
	pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);
	ioport = pinfo->ioport;

	
// Check transmitter for any frames that have been 
// sent and show the status.
	

	// Now check reciever
	nextRx = pinfo->nextRx;
//	DEBUG5(("rcv: CSR0  = %x\n", _inpw(ioport+RDP))); // interrupt mode

// If owned by the adapter, then no frame received yet.
	
	rmd1 = pinfo->rxDesc[nextRx].rmd1;
	if (rmd1 & RMD1_OWN) {
		return(AMERROR_OK);
	}


	DEBUG5(("Frame received.  Len: %d, Desc #: %d, Rmd1: %04x\n",
		pinfo->rxDesc[nextRx].rmd3 & RMD_CNT,
		nextRx,
		rmd1));

// If the frame doesn't have both STRT and END, then it's too
// big or invalid.  Kill frame and return.
	
	if ((rmd1 & (RMD1_STP | RMD1_ENP)) != (RMD1_STP | RMD1_ENP)) {
		DEBUG3(("STP and ENP not both set: %04x\n", rmd1));
		ret = AMERROR_INVALID_FRAME;
		goto RECEIVE_CLEANUP;
	}
	
// Check length of received frame
	
	rxLength = pinfo->rxDesc[nextRx].rmd3 & RMD_CNT;
	if (rxLength > buffLength) {
		DEBUG3(("Frame too big\n"));
		ret = AMERROR_TOO_BIG;
		goto RECEIVE_CLEANUP;
	}
	
// Check for errors
	
	if(pinfo->rxDesc[nextRx].rmd1 & RMD1_ERR)
	{
		DEBUG5(("Frame error rmd1 = %x\n",pinfo->rxDesc[nextRx].rmd1));
		ret = AMERROR_INVALID_FRAME;
		goto RECEIVE_CLEANUP;
	}
    
// Copy the data
	
	memcpy(rxBuff, pinfo->rxBuff[nextRx], rxLength);
	*prxLength = rxLength;
	
#if DEBUG_LEVEL > 5
	
// Dump the receive buffer
	
	for (ii = 0; ii < rxLength; ii += 20) {
		dptr = dbuff;
		for (jj = ii; (jj < ii + 20) && (jj < rxLength); jj++) {
			sprintf(dptr, "%02x ", rxBuff[jj]);
			dptr += 3;
		}
		DEBUG5(("Rx: %s\n", dbuff));
	}

#endif

RECEIVE_CLEANUP:
	_outpw(ioport+RAP, 0);   			// set to CSR0
	_outpw(ioport+RDP, _inpw(ioport+RDP)|CSR0_RINT);	// clear out RINT
	pinfo->rxDesc[nextRx].rmd3 = 0;		//clear the byte count
	pinfo->rxDesc[nextRx].rmd2 = RMD2_ONES|((-buffLength)&RMD_CNT );	//set the buffer byte count
	pinfo->rxDesc[nextRx].rmd1 = (rmd1&RMD1_ADDRH)|RMD1_OWN;  			//set the own bit to give
																		// this one back to the controller

	nextRx = (nextRx + 1) % PCN_MAX_RX;
	pinfo->nextRx = nextRx;
	return( ret );
}

/*********************************************************************
pcnTProcess()


Check the receiver and copy a TX buffer if any available.

RETURNS:
	AMERROR_OK		No frame, no problem.
	AMERROR_FRAME		Frame is transmitted.
	AMERROR_INVALID_UNIT	Invalid unit number.

*********************************************************************/

AMERROR
pcnTProcess(int unitNumber)
{
	int	readTx;
	PcnInfo_s *pinfo;
        int	ioport;

	if (unitNumber >= PCN_MAX) {
		DEBUG2(("Invalid unit number: %d\n", unitNumber));
		return(AMERROR_INVALID_UNIT);
	}
	pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);
	ioport = pinfo->ioport;

// Check transmitter for any frames that have been 
// sent and show the status.
	
	readTx = pinfo->readTx;
	if (readTx != pinfo->nextTx) 
		// If owned by adapter, not sent yet 
		
		if ((pinfo->txDesc[readTx].tmd1 & TMD1_OWN)==0)
		{
			DEBUG5(("Last Tx TMD1: %04x, TMD2: %04x, readTx: %d\n",
				pinfo->txDesc[readTx].tmd1,
				pinfo->txDesc[readTx].tmd2,readTx));
			readTx = (readTx + 1) % PCN_MAX_TX;
			pinfo->readTx = readTx;
			_outpw(ioport+RAP, 0);   	// set to CSR0
			_outpw(ioport+RDP, _inpw(ioport+RDP)|CSR0_TINT); // clear out RINT
			return(AMERROR_FRAME);
		}


	return( AMERROR_OK );
}



/*********************************************************************
pcnDump()


Check the receiver and copy a TX buffer if any available.

RETURNS:
		nothing.
*********************************************************************/

AMERROR
pcnDump(int unitNumber)
{
	PcnInfo_s *pinfo;
	int	ii;
    int	ioport;

	if (unitNumber >= PCN_MAX) {
		DEBUG2(("Invalid unit number: %d\n", unitNumber));
		return(AMERROR_INVALID_UNIT);
	}
	pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);
	ioport = pinfo->ioport;

	_outpw(ioport+RAP, 0);	
	DEBUG3(("CSR0: %04x\n", _inpw(ioport+RDP)));

	_outpw(ioport+RDP, CSR0_STOP);

#define PCSR(a) _outpw(ioport+RAP,a); DEBUG3(("CSR%d = %04x\n",a, _inpw(ioport+RDP)))
#define DCSR(a) _outpw(ioport+RAP,a); ii = _inpw(ioport+RDP); _outpw(ioport+RAP,a+1);\
		DEBUG3(("CSR%d,%d = %04x %04x\n",a, a+1,\
		  _inpw(ioport+RDP),ii))

	DCSR(0);
	PCSR(0);
	
	DCSR(2);
	DCSR(4);
	PCSR(6);
	PCSR(15);
	DCSR(18);
	DCSR(22);
	DCSR(24);
	DCSR(26);
	DCSR(28);
	DCSR(62);
	DCSR(66);
	DCSR(88);
	PCSR(40);
	PCSR(41);
	PCSR(44);
	PCSR(58);

//	_outpw(ioport+RAP, 18);
//	DEBUG3(("BCR18 = %04x\n", _inpw(ioport+IDP)));

//	pcnInitializeDescriptors(unitNumber);

	_outpw(ioport+RAP, 0);
	_outpw(ioport+RDP,CSR0_STRT);

}

/*********************************************************************
pcnInitializeDescriptors ()

set up the transmit and receive descriptors

RETURNS:
RETURNS:
	AMERROR_OK		No frame, no problem.
	AMERROR_FRAME		Frame is transmitted.
	AMERROR_INVALID_UNIT	Invalid unit number.

*********************************************************************/

AMERROR pcnInitializeDescriptors(int unitNumber)
{
	PcnInfo_s *pinfo;
	int	ii;
	unsigned long	faraddr;

	if (unitNumber >= PCN_MAX) {
		DEBUG2(("Invalid unit number: %d\n", unitNumber));
		return(AMERROR_INVALID_UNIT);
	}
	pinfo = pcnInfo + unitNumber*sizeof(PcnInfo_s);

// Set up the descriptors
	
	for (ii = 0; ii < PCN_MAX_RX; ii++) {
		FTP( &pinfo->rxBuff[ ii ][ 0 ],faraddr);
		pinfo->rxDesc[ii].rbadr = (VU16)(0xffff & faraddr);
		pinfo->rxDesc[ii].rmd1 = RMD1_OWN + (VU16)((0xff0000 & faraddr)>>16);
		pinfo->rxDesc[ii].rmd2 = (VU16)(RMD2_ONES|((-PCN_MAX_RX_SIZE)&RMD_CNT ));//set the buffer byte count
		pinfo->rxDesc[ii].rmd3 = 0;		//clear the byte count
		DEBUG5(("rxDesc[%d].rbadr = %04x, rmd1 = %04x\n", ii,
			pinfo->rxDesc[ii].rbadr,
			pinfo->rxDesc[ii].rmd1));
	}
               
	for (ii = 0; ii < PCN_MAX_TX; ii++) {
		FTP( &pinfo->txBuff [ ii ] [ 0 ],faraddr);
		pinfo->txDesc[ii].tbadr = (VU16)(0xffff & faraddr);
		pinfo->txDesc[ii].tmd1 = (VU16)((0xff0000 & faraddr)>>16);
	}

	pinfo->nextTx	= 0;
	pinfo->readTx	= 0;
	pinfo->nextRx   = 0;
	return(AMERROR_OK);
}

// Table to convert from PCN_MAX_nX to the value in tlen and rlen
// A 0 terminates the table and indicates an invalid PCN_MAX_nX value

const struct xLenTab_s {
	int	max_nX;
	int	xlen;
} xLenTab[] = {
	1,	0,
	2,	1,
	4,	2,
	8,	3,
	16,	4,
	32,	5,
	64,	6,
	128,	7,
	256,	8,
	512,	9,
	0,0
};


/*********************************************************************
Convert from the number of descriptors to the
value to be used as tlen and rlen.

RETURNS:
	AMERROR_OK		Value found *xlen contains proper value.
	AMERROR_NOTFOUND	Invalid PCN_MAX_nX
*********************************************************************/


AMERROR
xLenConv(int max_nX, int *xlen)
{
	int ii;

	for (ii = 0; xLenTab[ii].max_nX; ii++) {
		if (xLenTab[ii].max_nX == max_nX) {
			*xlen = xLenTab[ii].xlen;
			return(AMERROR_OK);
		}
	}
	return(AMERROR_NOT_FOUND);
}

